
#ifndef MPROGDEFS_H
#define MPROGDEFS_H

/* file generated by qcc, do not modify */

/*
typedef struct m_globalvars_s
{
	int	pad[28];
	int	self;
} m_globalvars_t;

typedef struct m_entvars_s
{
} m_entvars_t;

#define M_PROGHEADER_CRC 10020

*/

#endif
