/* file generated by qcc, do not modify */


#ifndef CLPROGDEFS_H
#define CLPROGDEFS_H

/*
typedef struct cl_globalvars_s
{
	int			pad[28];
	int			self;
	int			other;
	int			world;
	float		time;
	float		frametime;
	float		player_localentnum;
	float		player_localnum;
	float		maxclients;
	float		clientcommandframe;
	float		servercommandframe;
	string_t	mapname;
	vec3_t		v_forward;
	vec3_t		v_up;
	vec3_t		v_right;
	float		trace_allsolid;
	float		trace_startsolid;
	float		trace_fraction;
	vec3_t		trace_endpos;
	vec3_t		trace_plane_normal;
	float		trace_plane_dist;
	int			trace_ent;
	float		trace_inopen;
	float		trace_inwater;
	func_t		CSQC_Init;
	func_t		CSQC_Shutdown;
	func_t		CSQC_InputEvent;
	func_t		CSQC_UpdateView;
	func_t		CSQC_ConsoleCommand;
	vec3_t		pmove_org;
	vec3_t		pmove_vel;
	vec3_t		pmove_mins;
	vec3_t		pmove_maxs;
	float		input_timelength;
	vec3_t		input_angles;
	vec3_t		input_movevalues;
	float		input_buttons;
	float		movevar_gravity;
	float		movevar_stopspeed;
	float		movevar_maxspeed;
	float		movevar_spectatormaxspeed;
	float		movevar_accelerate;
	float		movevar_airaccelerate;
	float		movevar_wateraccelerate;
	float		movevar_friction;
	float		movevar_waterfriction;
	float		movevar_entgravity;
} cl_globalvars_t;

typedef struct cl_entvars_s
{
	float		modelindex;
	vec3_t		absmin;
	vec3_t		absmax;
	float			entnum;
	float			drawmask;
	func_t			predraw;
	float		movetype;
	float		solid;
	vec3_t		origin;
	vec3_t		oldorigin;
	vec3_t		velocity;
	vec3_t		angles;
	vec3_t		avelocity;
	string_t	classname;
	string_t	model;
	float		frame;
	float		skin;
	float		effects;
	vec3_t		mins;
	vec3_t		maxs;
	vec3_t		size;
	func_t		touch;
	func_t		use;
	func_t		think;
	func_t		blocked;
	float		nextthink;
	int			chain;
	string_t	netname;
	int			enemy;
	float		flags;
	float		colormap;
	int			owner;
} cl_entvars_t;

#define CL_PROGHEADER_CRC 52195
*/

#endif
