#include <windows.h> // include for version info constants

107 ICON "darkplaces.ico"

1 VERSIONINFO
FILEVERSION 1,0,0,0
PRODUCTVERSION 1,0,0,0
FILETYPE VFT_APP
{
  BLOCK "StringFileInfo"
	 {
		 BLOCK "040904E4"
		 {
			 VALUE "CompanyName", "DarkPlaces Contributors"
			 VALUE "FileVersion", "1.0"
			 VALUE "FileDescription", "DarkPlaces Game Engine"
			 VALUE "InternalName", "darkplaces.exe"
			 VALUE "LegalCopyright", "id Software, Forest Hale, and contributors"
			 VALUE "LegalTrademarks", ""
			 VALUE "OriginalFilename", "darkplaces.exe"
			 VALUE "ProductName", "DarkPlaces"
			 VALUE "ProductVersion", "1.0"
		 }
	 }
}
